"""A module for the conversion of units."""
import copy
import functools
import itertools
import os
import re
from typing import Union
import warnings

import numpy as np
import pandas as pd

from carpy.utility._miscellaneous import cast2numpy, GetPath, Hint
from carpy.utility._vanity import Unicodify

__all__ = ["Quantity", "cast2quantity"]
__author__ = "Yaseen Reza"


# ============================================================================ #
# Load data required by the module
# ---------------------------------------------------------------------------- #
def load_quantities_dfs():
    """Load and return the quantity dimensions and prefixes spreadsheets."""
    filepath = os.path.join(GetPath.localpackage(), "data", "quantities.xlsx")
    dataframes = pd.read_excel(io=filepath, sheet_name=None, na_filter=False)
    return dataframes


dfs = load_quantities_dfs()

# ============================================================================ #
# Unit Comprehension
# ---------------------------------------------------------------------------- #

# Patterns need to be sorted longest element to shortest element for regex
# This pattern collects all the possible prefixes carpy will recognise and parse
pattern_prefixes = "|".join([
    f"(?:{sym})" for sym in sorted(dfs["prefixes"]["Symbol"], key=len)[::-1]])


class Dimensions(object):
    """
    This class can be used to create and track the dimensionality of any
    supported units, and find the dimensionally equivalent SI-base counterpart.

    """
    # Order base units as per the MKS (metre, kilogram, second) system of units
    _sibase = np.array(["kg", "m", "s", "A", "K", "mol", "cd"])

    # This cache is *very* important to performance. pandas.series lookups in
    # this class are very slow. By saving __init__ outcomes to a cache, some
    # downstream functions see a 5-7x speed improvement! This cache does depend
    # on users sticking to a regular way of writing units though, for example,
    # 'm s^{-1}' and 's^{-1} m' are not equivalent. The hope is that users do
    # stick a regular way of writing units (due to inherent user preferences),
    # which may even coincide with units used by the library.
    _sessioncache = {}

    def __init__(self, units: str = None, /):
        """
        Args:
            units: A space delimited string designating the units to instantiate
                this object with. If the unit symbol is ambiguous, the unit
                should be followed by '.<unit-system>' i.e. .SI, .metric, .USC,
                or .imperial tags.
        """
        self._unitstring = "" if units is None else units

        # Check cache for an easy answer
        # (use deepcopy because of dictionary memory persistence nonsense)
        if self._unitstring in self._sessioncache:
            self._pairings = copy.deepcopy(self._sessioncache[self._unitstring])
            return

        # Otherwise, break units into components and proceed
        units = self._unitstring.split()

        # Dictionary will collapse repeated units into singular elements
        pairings = dict(zip(units, [dict() for _ in units]))
        for i, key in enumerate(units):

            unit = units[i]

            # Assess the power of the unit's dimension, consuming the string
            if not pairings[key].get("power"):
                pairings[key]["power"] = 0.0
            # Have to iadd the power in case there was a similar term earlier
            powerterm = re.findall(r"\^{(-?[0-9.]+)}", unit)
            if powerterm:
                pairings[key]["power"] += float(powerterm[0])
                unit = unit[:unit.index("^")]
            else:
                pairings[key]["power"] += 1.0

            # If the unit's system is given, determine it here
            system = re.findall(r"\.(.+)", unit)
            if system:
                system, = system  # Unpack findall's list into a string
                unit = unit[:unit.index(".")]
                # Check that this unit doesn't already have a dissimilar system
                if pairings[key].get("system", system) != system:
                    errormsg = (
                        f"'{unit}' in '{self._unitstring}' has conflicting unit"
                        f" systems. Deconflict by explicitly stating the system"
                        f" of this unit"
                    )
                    raise ValueError(errormsg)
                pairings[key]["system"] = system

            # Look for prefixes in the unit
            df_prefix = dfs["prefixes"]
            pattern = f"^(?:{pattern_prefixes})"
            prefixes = dict(
                df_prefix[
                    np.array([
                        x in set([""] + re.findall(pattern, unit))
                        for (i, x) in enumerate(df_prefix["Symbol"])
                    ])
                ]["Symbol"]
            )
            # If the prefix belongs to the Bel system, only Bel symbols allowed
            if "dB" in prefixes.values():
                pairings[key]["system"] = "Bel"

            # Create a dictionary of symbols indexed by the dataframe row number
            df_dims = dfs["dimensions"]
            system = pairings[key].get("system")
            symbols = df_dims["Symbol(s)"]
            if system is not None:
                symbols = dict(symbols[df_dims["System"] == system])
            else:
                symbols = dict(symbols[df_dims["System"] != "Bel"])

            # Try to pair a prefix with a unit symbol that completes the unit
            products = [
                (prefix_id, symbol_id)
                for prefix_id, symbol_id in itertools.product(prefixes, symbols)
                if unit in [
                    f"{prefixes[prefix_id]}{x}"
                    for x in symbols[symbol_id].split(",")
                ] and not (  # Fail condition: SI prefix on non-SI unit
                        df_prefix["System"][prefix_id] == "SI"
                        and prefixes[prefix_id] != ""
                        and df_dims["System"][symbol_id] != "SI"
                )
            ]
            if len(products) > 1:
                matched = [
                    f"<{symbols[i]}>.{df_dims['System'][i]}"
                    for _, i in products
                ]
                errormsg = (
                    f"'{key}' is too ambiguous a declaration of a unit. "
                    f"Discovered potential matches to symbols: {matched}"
                )
                raise ValueError(errormsg)
            elif len(products) == 0:
                errormsg = (
                    f"Could not identify what unit '{key}' might be referring "
                    f"to in the compound unit '{self._unitstring}'."
                )
                raise ValueError(errormsg)

            # Otherwise all is good, save the result...
            product, = products
            _, sym_id = product
            symbol = df_dims["Symbol(s)"][sym_id].split(",")[0]

            pairings[key]["(prefix, unit)"] = product
            pairings[key]["symbol"] = symbol
            pairings[key]["system"] = df_dims["System"][sym_id]

        # Store and write to cache
        self._pairings = pairings
        self._sessioncache[self._unitstring] = copy.deepcopy(self._pairings)
        return

    def __repr__(self):
        return f"{type(self).__name__}('{self._unitstring}')"

    def __str__(self):
        return Unicodify.mathscript_safe(self.si_equivalent)

    def __mul__(self, other):
        # Add dimensional powers
        dict_a = {**self.si_dimensioned, **self.si_dimensionless}
        dict_b = {**other.si_dimensioned, **other.si_dimensionless}
        dict_sum = {
            k: dict_a.get(k, 0) + dict_b.get(k, 0)
            for k in set(dict_a) | set(dict_b)
        }
        # Create a new instance and overwrite hidden attributes
        unitstring = " ".join(
            f"{k}.SI" + "^{%f}" % v for k, v in dict_sum.items()
        )
        output = self.__class__(unitstring)
        for k in output._pairings:
            output._pairings[k]["power"] = dict_sum[k[:k.index(".")]]
        return output

    def __truediv__(self, other):
        # Subtract dimensional powers
        dict_a = {**self.si_dimensioned, **self.si_dimensionless}
        dict_b = {**other.si_dimensioned, **other.si_dimensionless}
        dict_sum = {
            k: dict_a.get(k, 0) - dict_b.get(k, 0)
            for k in set(dict_a) | set(dict_b)
        }
        # Create a new instance and overwrite hidden attributes
        unitstring = " ".join(
            f"{k}.SI" + "^{%f}" % v for k, v in dict_sum.items())
        output = self.__class__(unitstring)
        for k in output._pairings:
            output._pairings[k]["power"] = dict_sum[k[:k.index(".")]]
        return output

    def __pow__(self, power, modulo=None):

        myunits = {**self.si_dimensioned, **self.si_dimensionless}

        # Create a new instance and overwrite hidden attributes
        unitstring = " ".join(
            f"{k}.SI" + "^{%f}" % v for k, v in myunits.items())
        output = self.__class__(unitstring)
        for k in output._pairings:
            output._pairings[k]["power"] = myunits[k[:k.index(".")]] * power

        return output

    def __invert__(self):
        # Simply return true division result
        return self.__class__("").__truediv__(self)

    @property
    def og_components(self) -> dict:
        """A dictionary describing this object's constituent units."""
        return self._pairings

    @property
    def og_string(self) -> str:
        """The original string that instantiated this unit object."""
        return self._unitstring

    @functools.cached_property
    def si_dimensioned(self) -> dict:
        """Return a dictionary dimensionality of the unit in SI base units."""
        dims = self._sibase
        mydims = dict(zip(dims, np.zeros(len(dims))))

        # Iterate over each component unit of the units object
        for _, (_, pairinfo) in enumerate(self._pairings.items()):
            _, unitid = pairinfo["(prefix, unit)"]
            unitpower = pairinfo["power"]
            unitseries = dfs["dimensions"].loc[unitid]

            # Iterate over each SI dimension and compound the answer
            for baseunit in dims:
                mydims[baseunit] += unitseries[baseunit] * unitpower

        return mydims

    @functools.cached_property
    def si_dimensionless(self) -> dict:
        """Return a dictionary of non-dimensional components of the unit."""
        dims = np.array(["rad", "sr"])
        mydims = dict(zip(dims, np.zeros(len(dims))))

        # Iterate over each component unit of the units object
        for _, (_, pairinfo) in enumerate(self._pairings.items()):
            _, unitid = pairinfo["(prefix, unit)"]
            unitpower = pairinfo["power"]
            unitseries = dfs["dimensions"].loc[unitid]

            # Iterate over each SI dimension and compound the answer
            for baseunit in dims:
                mydims[baseunit] += unitseries[baseunit] * unitpower

        return mydims

    @property
    def si_equivalent(self) -> str:
        """Return a string of dimensions in SI equivalent units."""
        stringelements = []
        si_dimensions = {**self.si_dimensionless, **self.si_dimensioned}
        for _, (symbol, power) in enumerate(si_dimensions.items()):
            if power == 1:
                stringelements.append(symbol)
            elif power == 0:
                pass
            elif int(power) == power:
                stringelements.append(symbol + "^{%d}" % power)
            else:
                stringelements.append(symbol + "^{%.3f}" % power)

        return " ".join(stringelements)


class Quantity(np.ndarray):
    """
    The Quantity class can be used to maintian consistency in the manipulation
    of quantities, measures, and their units.

    This class inherits many properties of classic numpy arrays.
    """

    def __new__(cls, value: Hint.nums, /, units: Union[str, Dimensions] = None):
        """
        Args:
            value: Any number(s) that you would expect a numpy array to support.
            units: Any argument the 'Units' class of this module supports.
        """
        # Recast as necessary
        value = cast2numpy(value, dtype=np.float64)
        units = units if isinstance(units, Dimensions) else Dimensions(units)

        # If the value is a quantity, raise error to user on recasted dimensions
        if isinstance(value, Quantity) and (
                value.units.si_equivalent != units.si_equivalent
        ):
            errormsg = f"Quantity {value} had dimensions recast to {units}"
            raise ValueError(errormsg)

        # Try and convert the value to SI values
        if units.og_string == "degC":
            value = value + 273.15
        elif units.og_string == "degF":
            value = (value - 32) * (5 / 9) + 273.15
        else:
            for _, (_, pairinfo) in enumerate(units.og_components.items()):
                prefix_id, unit_id = pairinfo["(prefix, unit)"]

                prefix_system = dfs["prefixes"]["System"][prefix_id]

                # Different conversions depending on the system of the prefix
                if prefix_system == "SI":
                    value = value * (
                            10.0 ** dfs["prefixes"]["Exponent"][prefix_id]
                            * dfs["dimensions"]["transform"][unit_id]
                    ) ** pairinfo["power"]

                elif prefix_system == "Bel":
                    value = value + dfs["dimensions"]["transform"][unit_id]
                    value = 10.0 ** (value / 10)  # map dB scale to linear scale

                else:
                    raise NotImplementedError(f"No method for {prefix_system}")

        # Now that values are in SI format, we want units of a new obj to be too
        # ... identify components of a unit string based on SI dimensions
        newunit_components = [
            f"{k}.SI" + ("^{%f}" % v)
            for k, v in units.si_dimensioned.items()
        ]
        # ... identify components of a unit string based on angular dimensions
        newunit_components += [
            f"{k}.SI" + ("^{%f}" % v)
            for k, v in units.si_dimensionless.items()
        ]
        newunit_str = " ".join(newunit_components)

        # The new instance of this class should be an array object
        customarray = np.asarray(value).view(cls)
        # Add a custom attribute to the array object
        customarray.units = Dimensions(newunit_str)
        return customarray

    def __array_finalize__(self, obj, **kwargs):
        # This magic method is required for numpy compatibility
        if obj is None:
            return
        self.units = getattr(obj, "units", None)

    def __repr__(self):
        return f"{super().__repr__()[:-1]}, units='{self.units}')"

    def __str__(self):
        si_units_utf8 = Unicodify.mathscript_safe(self.units.si_equivalent)
        if len(self.flat) != 1:
            return f"{super().__str__()} {si_units_utf8}"
        return f"{super().__str__()[1:-1]} {si_units_utf8}"

    # --------------------- #
    # Relational Operations #
    # --------------------- #

    def _can_do_relational(self, other: object):
        """Check that it makes sense to compare two objects as Quantity objs."""
        if isinstance(other, self.__class__):
            dims0 = {**self.units.si_dimensionless,
                     **self.units.si_dimensioned}
            dims1 = {**other.units.si_dimensionless,
                     **other.units.si_dimensioned}
            if not (dims0 == dims1):
                # Two quantity objects with mismatched dimensions
                errormsg = (
                    f"Cannot carry out operation between {type(self).__name__} "
                    f"objects with dimensions do not match (got "
                    f"{self.units} and {other.units})"
                )
                raise ValueError(errormsg)
        return True

    def __lt__(self, other):
        if self._can_do_relational(other):
            return np.array(self) < other

    def __le__(self, other):
        if self._can_do_relational(other):
            return np.array(self) <= other

    def __eq__(self, other):
        if self._can_do_relational(other):
            return np.array(self) == other

    def __ne__(self, other):
        if self._can_do_relational(other):
            return np.array(self) != other

    def __gt__(self, other):
        if self._can_do_relational(other):
            return np.array(self) > other

    def __ge__(self, other):
        if self._can_do_relational(other):
            return np.array(self) >= other

    # --------------------- #
    # Arithmetic Operations #
    # --------------------- #

    def __add__(self, other):
        if self._can_do_relational(other):  # <-- Ensures matching units
            return super().__add__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if self._can_do_relational(other):  # <-- Ensures matching units
            return super().__sub__(other)

    def __rsub__(self, other):
        return -self.__sub__(other)

    def __isub__(self, other):
        return self.__sub__(other)

    def __mul__(self, other):
        # For mismatched objects, use default behaviour
        if not isinstance(other, type(self)):
            return super().__mul__(other)

        # Otherwise 2x Quantity objects. Create and return a new object
        value = np.array(self) * np.array(other)
        units = self.units * other.units
        return self.__class__(value, units)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __matmul__(self, other):
        warnmsg = (
            f"{type(self).__name__} objects do not yet support the '@' "
            f"operator. Collapsing result into regular numpy arrays w/o units"
        )
        warnings.warn(message=warnmsg, category=RuntimeWarning)
        return np.array(self).__matmul__(other)

    def __rmatmul__(self, other):
        warnmsg = (
            f"{type(self).__name__} objects do not yet support the '@' "
            f"operator. Collapsing result into regular numpy arrays w/o units"
        )
        warnings.warn(message=warnmsg, category=RuntimeWarning)
        return np.array(self).__rmatmul__(other)

    def __imatmul__(self, other):
        warnmsg = (
            f"{type(self).__name__} objects do not yet support the '@' "
            f"operator. Collapsing result into regular numpy arrays w/o units"
        )
        warnings.warn(message=warnmsg, category=RuntimeWarning)
        return np.array(self).__imatmul__(other)

    def __truediv__(self, other):
        # For mismatched objects, use default behaviour
        if not isinstance(other, type(self)):
            return super().__truediv__(other)

        # Otherwise 2x Quantity objects. Create and return a new object
        value = np.array(self) / np.array(other)
        units = self.units / other.units
        return self.__class__(value, units)

    def __rtruediv__(self, other):
        # For mismatched objects, cast both to Quantity objects.
        if not isinstance(other, type(self)):
            other = Quantity(other, "")

        # 2x Quantity objects. Create and return a new object
        return other.__truediv__(self)

    def __itruediv__(self, other):
        return self.__truediv__(other)

    def __floordiv__(self, other):
        # For mismatched objects, use default behaviour
        if not isinstance(other, type(self)):
            return super().__floordiv__(other)

        # Otherwise 2x Quantity objects. Create and return a new object
        value = np.array(self) // np.array(other)
        units = self.units / other.units
        return self.__class__(value, units)

    def __rfloordiv__(self, other):
        # For mismatched objects, cast both to Quantity objects.
        if not isinstance(other, type(self)):
            other = Quantity(other, "")

        # 2x Quantity objects. Create and return a new object
        return other.__floordiv__(self)

    def __ifloordiv__(self, other):
        return self.__floordiv__(other)

    def __mod__(self, other):
        # For mismatched objects, use default behaviour
        if not isinstance(other, type(self)):
            return super().__mod__(other)

        # Otherwise 2x Quantity objects. Create and return a new object
        value = np.array(self) - (np.array(self) // np.array(other))
        units = self.units / other.units
        return self.__class__(value, units)

    def __rmod__(self, other):
        # For mismatched objects, cast both to Quantity objects.
        if not isinstance(other, type(self)):
            other = Quantity(other, "")

        # 2x Quantity objects. Create and return a new object
        return other.__mod__(self)

    def __imod__(self, other):
        return self.__mod__(other)

    def __divmod__(self, other):
        return self.__floordiv__(other), self.__mod__(other)

    def __rdivmod__(self, other):
        # For mismatched objects, cast both to Quantity objects.
        if not isinstance(other, type(self)):
            other = Quantity(other, "")

        # 2x Quantity objects. Create and return a new object
        return other.__divmod__(self)

    def __pow__(self, power, modulo=None):

        if modulo is not None:
            return NotImplemented

        if isinstance(power, self.__class__) and str(power.units) != "":
            errormsg = (
                "Exponent cannot be a Quantity object with non-negligible units"
            )
            raise ValueError(errormsg)

        value = np.array(super().__pow__(power, modulo))
        units = self.units ** power
        return Quantity(value, units)

    def __rpow__(self, power):

        if str(self.units) != "":
            errormsg = (
                "Exponent cannot be a Quantity object with non-negligible units"
            )
            raise ValueError(errormsg)
        return super().__rpow__(power)

    def __ipow__(self, power):
        return self.__pow__(power)

    # ---------------------- #
    # Arithmetic Overloading #
    # ---------------------- #

    def __round__(self, n=None):
        return super().round(decimals=n if n is not None else 0)

    def __trunc__(self):
        return np.trunc(self)

    def __floor__(self):
        return np.floor(self)

    def __ceil__(self):
        return np.ceil(self)

    def __float__(self):
        flattened = self.flatten()
        # If Quantity only contains a single scalar value, take it!
        if len(flattened) == 1:
            return float(self.flat[0])  # *actually* cast it as a float
        # Otherwise we're losing information unintentionally, raise an error
        else:
            errormsg = "Conversion of multi-element array to scalar is lossy"
            raise ValueError(errormsg)

    # ----------------- #
    # Slice Overloading #
    # ----------------- #
    def __getitem__(self, key):
        return Quantity(super().__getitem__(key), self.units)

    @property
    def x(self) -> Union[float, np.ndarray]:
        """The raw value(s) of the numpy array, without associated units."""
        try:
            return float(self)
        except:
            return np.array(self)

    @property
    def u(self) -> Dimensions:
        """Easy access property for this quantity object's units."""
        return self.units

    def to(self, units: str, /) -> np.ndarray:
        """
        Convert this quantity into a desired unit, provided that the units have
        compatible dimensions.

        Args:
            units: A space delimited string designating the units to instantiate
                this object with. If the unit symbol is ambiguous, the unit
                should be followed by '.<unit-system>' i.e. .SI, .metric, .USC,
                or .imperial tags.

        Returns:
            The value of this quantity in the specified units.

        """
        # Recast as necessary
        units = units if isinstance(units, Dimensions) else Dimensions(units)

        errormsg = f"Incompatible units ({self.units} != {units})"
        assert self.units.si_equivalent == units.si_equivalent, errormsg

        # Special cases where the unit needs to shift the scale
        if len(units.og_components) == 1:
            if units.og_string == "degC":
                return self.x - 273.15
            elif units.og_string == "degF":
                return (self.x - 273.15) * (9 / 5) + 32

            # Special cases where the scale is logarithmic
            elif units.og_string.startswith("dB"):
                pairinfo = units.og_components[units.og_string]
                _, unit_id = pairinfo["(prefix, unit)"]
                transform = dfs["dimensions"]["transform"][unit_id]
                return 10 * np.log10(self.x) - transform
        else:
            for _, (_, pairinfo) in enumerate(units.og_components.items()):
                if pairinfo["system"] == "Bel":
                    raise ValueError("(deci)Bel conversion not recognised")

        # Converting to units which differ by a scalar multiple is as easy as...
        output = self / self.__class__(1, units)

        # If it was successful, the si_equivalent units should've cancelled
        if output.units.si_equivalent == "":
            return output.x

        # The units weren't successfully mapped due to inconsistent units
        errormsg = (
            f"Target '{units=}' do not share dimensionality with {self.units=}")
        raise ValueError(errormsg)


def cast2quantity(pandas_df: pd.DataFrame) -> dict:
    """
    Convert the contents of a dataframe into a dictionary of quantities.

    Args:
        pandas_df: A pandas dataframe object.

    Returns:
        A dictionary of quantities.

    """
    # Fill blank values with NaN
    pandas_df = pandas_df.mask(pandas_df == "")

    # Iterate over the columns in the dataframe
    output = dict()
    for column_name in pandas_df.columns:
        name_components = re.split(r"\s\[(.+)]", column_name)
        new_values = pandas_df[column_name]

        # If units are not present, just copy the values over to the output
        if len(name_components) == 1:
            output[column_name] = new_values
            continue

        # Otherwise, check that the column doesn't already exist in output
        output_col, units, _ = name_components
        if output_col not in output:
            output[output_col] = Quantity(new_values, units)

        # If column name already exists, assume both new column and old column
        # are referring to the same quantity - just different units. Try to
        # update missing elements of the original column with the new column
        else:
            wherenan = np.isnan(output[output_col])
            output[output_col][wherenan] = Quantity(new_values, units)[wherenan]

    return output



import unittest

from carpy.structures import ContinuousIndex, DiscreteIndex


class IndexTemplates(unittest.TestCase):
    """Test the template parent classes for cross-sectional stations."""

    def test_continuous_indexing(self):
        """Station data is defined with a continuous function."""
        stations = ContinuousIndex(lambda x: x * 2)
        self.assertEqual(stations[3], 6)
        return

    def test_discrete_indexing(self):
        """Station data is defined with discrete elements"""
        # Initial assignment
        stations = DiscreteIndex({1: 100, -2: -50})

        # Brand-new assignment
        stations[0] = 0

        # Update assignment
        stations[-2] = -200

        # Singular keys
        self.assertEqual(stations[-2], -200)
        self.assertEqual(stations[0], 0)
        self.assertEqual(stations[1], 100)

        # Real slices
        self.assertEqual(stations[0:], [0, 100])  # Start only
        self.assertEqual(stations[:0], [-200])  # Stop only
        self.assertEqual(stations[:0.5], [-200, 0])  # Stop only, new index
        self.assertEqual(stations[:], [-200, 0, 100])  # All defined indices

        # Interpolated keys
        self.assertEqual(stations[0.5], 50)

        # Interpolated slices
        self.assertEqual(stations[::4], [-200, -100, 0, 100])
        return


if __name__ == '__main__':
    unittest.main()



"""Tests for weather tools in carpy."""
import unittest

from carpy.environment import METARtools


class METARtoolkit(unittest.TestCase):
    """Tests for METAR parsing and fetching tools in carpy."""

    def test_fetching(self):
        """Test if we can fetch METAR data."""
        metars_egll = METARtools.fetch("EGLL")
        self.assertTrue(isinstance(metars_egll[0], str))
        return



"""Module implementing methods for compressible flow calculations."""
from functools import partial, cached_property
import warnings

import numpy as np
from scipy.optimize import newton

from carpy.utility import Hint, cast2numpy, constants as co, Quantity, isNone

__all__ = ["V_max", "nu_max", "IsentropicFlow", "NormalShock", "ObliqueShock",
           "ExpansionFan", "RayleighFlow", "FannoFlow"]
__author__ = "Yaseen Reza"


def set_gamma(gamma: Hint.nums = None, /) -> np.ndarray:
    """Return default sea-level gamma if it is not provided."""
    gamma = co.STANDARD.SL.gamma if gamma is None else gamma
    return np.array(cast2numpy(gamma))


def set_beta(beta: Hint.nums = None) -> np.ndarray:
    """Return Mach angle default beta=pi/2 if it is not provided."""
    beta = beta if beta is not None else np.pi / 2
    return cast2numpy(beta)


def set_theta(theta: Hint.nums = None) -> np.ndarray:
    """Return deflection angle default theta=0 if it is not provided."""
    theta = theta if theta is not None else 0
    return cast2numpy(theta)


def V_max(cp: Hint.nums, Ttotal: Hint.nums):
    """
    Maximum attainable velocity of a gas as a function of energy content.

    Args:
        cp: Specific heat of the gas at constant pressure.
        Ttotal: Total (stagnation) temperature of the flow.

    Returns:
        Maximum thermodynamic velocity of the gas.

    """
    # Recast as necessary
    cp = cast2numpy(cp)
    Ttotal = cast2numpy(Ttotal
                        )
    maximum_velocity = np.sqrt(2 * cp * Ttotal)
    return maximum_velocity


def nu_max(gamma: Hint.nums = None):
    """
    Maximum permissible Prandtl-Meyer angle of a flow.

    Args:
        gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

    Returns:
        The limiting Prandtl-Meyer angle.

    """
    # Recast as necessary
    gamma = set_gamma(gamma)

    # Limit of Prandtl-Meyer angle (Mach inf.)
    maximum_angle = np.pi / 2 * (((gamma + 1) / (gamma - 1)) ** 0.5 - 1)
    return maximum_angle


class IsentropicFlow(object):
    """
    Classic one- and two-dimensional isentropic compressible flow methods for
    an ideal gas.
    """

    @staticmethod
    def T_T0(M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the static to total temperature ratio.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The static to total temperature ratio.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Isentropic relation
        ratio = (1 + (gamma - 1) / 2 * M ** 2) ** -1

        return ratio

    @classmethod
    def T_Tstar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the static to sonic temperature ratio.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The static to sonic temperature ratio.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute sonic ratio
        T_Ttotal = cls.T_T0(M=M, gamma=gamma)
        Tstar_Ttotal = cls.T_T0(M=1, gamma=gamma)
        ratio = T_Ttotal / Tstar_Ttotal

        return ratio

    @classmethod
    def p_p0(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the static to total pressure ratio.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The static to total pressure ratio.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Isentropic relation
        ratio = cls.T_T0(M=M, gamma=gamma) ** (gamma / (gamma - 1))

        return ratio

    @classmethod
    def p_pstar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the static to sonic pressure ratio.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The static to sonic pressure ratio.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute sonic ratio
        p_ptotal = cls.p_p0(M=M, gamma=gamma)
        pstar_ptotal = cls.p_p0(M=1, gamma=gamma)
        ratio = p_ptotal / pstar_ptotal

        return ratio

    @classmethod
    def rho_rho0(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the static to total density ratio.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The static to total density ratio.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Isentropic relation
        ratio = cls.T_T0(M=M, gamma=gamma) ** (1 / (gamma - 1))

        return ratio

    @classmethod
    def rho_rhostar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the static to sonic density ratio.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The static to sonic density ratio.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute sonic ratio
        rho_rhototal = cls.rho_rho0(M=M, gamma=gamma)
        rhostar_rhototal = cls.rho_rho0(M=1, gamma=gamma)
        ratio = rho_rhototal / rhostar_rhototal

        return ratio

    @classmethod
    def A_Astar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the static to sonic area ratio.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The static to sonic area ratio.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute terms
        M = np.where(M == 0, np.nan, M)
        term1 = ((gamma + 1) / 2) ** -((gamma + 1) / 2 / (gamma - 1))
        term2 = 1 / M / cls.rho_rho0(M=M, gamma=gamma) ** ((gamma + 1) / 2)

        # Compute sonic ratio
        ratio = term1 * term2

        return ratio

    @classmethod
    def mdot(cls, pt: Hint.nums, Tt: Hint.nums, Rs: Hint.nums, A: Hint.nums,
             M: Hint.nums, gamma: Hint.nums = None) -> Quantity:
        """
        Mass flow rate of a given station in an isentropic duct.

        Args:
            pt: Total pressure.
            Tt: Total temperature.
            Rs: Specific gas constant.
            A: Flow area.
            M: Mach number of flow.
            gamma: Adiabatic index of flow.

        Returns:
            The mass flow rate.

        """
        # Recast as necessary
        kwargs = cast2numpy({"pt": pt, "Tt": Tt, "Rs": Rs, "A": A})
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute choked mass flow rate
        mdot = kwargs["A"] * kwargs["pt"] * kwargs["Tt"] ** -0.5
        mdot = mdot * M * (gamma / kwargs["Rs"]) ** 0.5
        mdot = mdot * cls.rho_rho0(M=M, gamma=gamma) ** ((gamma + 1) / 2)

        return Quantity(mdot, "kg s^{-1}")

    @staticmethod
    def mu(M: Hint.nums) -> np.ndarray:
        """
        Compute the Mach angle.

        Args:
            M: Mach number of the flow.

        Returns:
            The flow Mach angle for M >=1 (giving mu <= pi/2), otherwise NaN.

        """
        # Recast as necessary
        M = cast2numpy(M)

        # Invalidate any result where Mach number is below 1
        M = np.where(M < 1, np.nan, M)

        # Compute angle
        mu = np.arcsin(1 / M)

        return mu

    @staticmethod
    def nu(M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the Prandtl-Meyer angle.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The Prandtl-Meyer angle for M >=1 (giving nu >= 0), otherwise NaN.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Invalidate any result where Mach number is below 1
        M = np.where(M < 1, np.nan, M)

        # Compute terms
        term1 = np.sqrt((gamma + 1) / (gamma - 1))
        term2 = np.sqrt(M ** 2 - 1)

        # Compute angle
        nu = term1 * np.arctan(term2 / term1) - np.arctan(term2)

        return nu

    @staticmethod
    def M(T_T0: Hint.nums = None, p_p0: Hint.nums = None,
          rho_rho0: Hint.nums = None, gamma: Hint.nums = None):
        """
        Compute the flow Mach number.

        Only one of the temperature, pressure, and density ratios need be
        specified.

        Args:
            T_T0: The static to total temperature ratio.
            p_p0: The static to total pressure ratio.
            rho_rho0: The static to total density ratio.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The flow Mach number as a function of one of the ratios and gamma.

        """
        # Recast as necessary
        # Verify validity of inputs, and that only one input is given
        if tuple(isNone(T_T0, p_p0, rho_rho0)).count(True) == 2:
            # Temperature ratio is given
            if p_p0 is None and rho_rho0 is None:
                pass
            # Pressure ratio is given
            elif T_T0 is None and rho_rho0 is None:
                T_T0 = p_p0 ** ((gamma - 1) / gamma)
            # Density ratio is given
            elif T_T0 is None and p_p0 is None:
                T_T0 = rho_rho0 ** (gamma - 1)
        else:
            errormsg = (
                f"Expected one of T_T0, p_p0, and rho_rho0 arguments should be "
                f"used (got {T_T0=}, {p_p0=}, {rho_rho0=})"
            )
            raise ValueError(errormsg)
        gamma = set_gamma(gamma)

        M = ((1 / T_T0 - 1) * (2 / (gamma - 1))) ** 0.5

        return M


class NormalShock(object):
    """
    Classic Rankine-Hugoniot one-dimensional shock methods for an ideal gas.
    """

    @staticmethod
    def p2_p1(M1: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the downstream over upstream static pressure ratio.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            Pressure ratio over the shock, p2/p1.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)

        # Compute ratio
        M1 = np.where(M1 < 1, np.nan, M1)
        ratio = (2 * gamma * M1 ** 2 - (gamma - 1)) / (gamma + 1)

        return ratio

    @staticmethod
    def rho2_rho1(M1: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the downstream over upstream static density ratio.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            Density ratio over the shock, rho2/rho1.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)

        # Compute ratio
        M1 = np.where(M1 < 1, np.nan, M1)
        ratio = (gamma + 1) * M1 ** 2 / ((gamma - 1) * M1 ** 2 + 2)

        return ratio

    @classmethod
    def V2_V1(cls, M1: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the downstream over upstream fluid velocity ratio.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            Fluid velocity ratio over the shock, U2/U1.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)

        # Compute ratio
        ratio = 1 / cls.rho2_rho1(M1=M1, gamma=gamma)

        return ratio

    @classmethod
    def T2_T1(cls, M1: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the downstream over upstream static temperature ratio.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            Temperature ratio over the shock, T2/T1.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)

        # Compute terms
        term1 = cls.p2_p1(M1=M1, gamma=gamma)
        term2 = cls.rho2_rho1(M1=M1, gamma=gamma)

        # Compute ratio
        ratio = term1 / term2

        return ratio

    @classmethod
    def p02_p01(cls, M1: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the downstream over upstream total pressure ratio.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            Total pressure ratio over the shock, pt2/pt1.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)

        # Compute terms
        term1 = cls.rho2_rho1(M1=M1, gamma=gamma)
        term2 = cls.p2_p1(M1=M1, gamma=gamma)

        # Compute ratio
        ratio = term1 ** (gamma / (gamma - 1)) * term2 ** -(1 / (gamma - 1))

        return ratio

    @staticmethod
    def T02_T01(*args, **kwargs) -> np.ndarray:
        """
        Compute the downstream over upstream total temperature ratio.

        Returns:
            Total temperature ratio over the shock, Tt2/Tt1(= 1).

        """
        if len(args) > 0 or len(kwargs) > 0:
            warnmsg = (
                f"Total temperature across a shock doesn't change, no need to "
                f"provide this method with any arguments"
            )
            warnings.warn(message=warnmsg, category=RuntimeWarning)
        return np.ones(1)

    @classmethod
    def M2(cls, M1: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the downstream Mach number.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            Mach number after the shock, M2.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)

        # Compute terms
        term1 = cls.rho2_rho1(M1=M1, gamma=gamma)
        term2 = cls.p2_p1(M1=M1, gamma=gamma)

        # Compute Mach
        M2 = ((M1 ** 2) / term1 / term2) ** 0.5

        return M2


class ObliqueShock(object):
    """
    Classic two-dimensional oblique shock functions for an ideal gas.
    """

    @staticmethod
    def theta(M1: Hint.nums, gamma: Hint.nums = None,
              beta: Hint.nums = None) -> np.ndarray:
        """
        Compute the resulting deflection angle of a flow after passing through
        an oblique shock that makes the angle beta with the flow upstream of the
        shock (beta = 90 degrees therefore indicates a normal shock).

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.
            beta: The acute angle the shockwave makes with the upstream flow.
                Optional, defaults to pi/2 (recovering normal shock relation).

        Returns:
            The angle with which the resulting flow is deflected.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)
        beta = set_beta(beta)

        # Compute terms
        term1 = np.tan(beta)
        term2 = (gamma + 1) * M1 ** 2
        term3 = 2 * (M1 ** 2 * np.sin(beta) ** 2 - 1)

        # Compute deflection angle
        theta = np.arctan(1 / (term1 * (term2 / term3 - 1)))

        return theta

    @classmethod
    def beta(cls, M1: Hint.nums, gamma: Hint.nums = None,
             theta: Hint.nums = None) -> tuple:
        """
        Compute the weak and strong shock angles that result from the turning of
        supersonic flow. In many cases, it is the weak shock result that is
        desirable for engineers.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.
            theta: The deflection angle of the downstream flow from upstream.
                Optional, defaults to 0 (recovering normal shock relation).

        Returns:
            tuple: Weak and strong shock solutions, respectively.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)
        theta = set_theta(theta)

        # Create a solver that we can vectorize
        def solver(a, b, c):
            """
            Helper function to solve the theta-beta-Mach problem.

            Args:
                a: Upstream Mach number.
                b: Upstream gamma.
                c: Deflection angle of the flow.

            Returns:
                tuple: weak and strong shock solutions, respectively.

            """

            def f_opt(x):
                """Helper function for root-finding."""
                # Find where theta = f(M1, beta) results in desired theta
                result = partial(cls.theta, M1=a, gamma=b)(beta=x) - c
                return result

            try:
                # Try to compute weak shock result
                beta_weakshock = newton(f_opt, np.pi / 180)

                # M=2, gamma=1.3, theta=28 deg gives a weird beta = 200 rad case
                if not (0 <= beta_weakshock <= np.pi / 2):
                    raise RuntimeError

            except RuntimeError as _:
                # Failure to compute suggests the shock has detached (bow shock)
                beta_weakshock = np.nan
                beta_strongshock = np.nan
            else:
                # Otherwise computation was fine, strong solution exists too
                beta_strongshock = newton(f_opt, np.pi / 2)

            return beta_weakshock, beta_strongshock

        # Return solution as a tuple of weak and strong solutions
        return np.vectorize(solver)(a=M1, b=gamma, c=theta)

    @classmethod
    def p2_p1(cls, M1: Hint.nums, gamma: Hint.nums = None,
              theta: Hint.nums = None) -> tuple:
        """
        Compute the downstream over upstream static pressure ratio.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.
            theta: The deflection angle of the downstream flow from upstream.
                Optional, defaults to 0 (recovering normal shock relation).

        Returns:
            weak shock solution, strong shock solution : Pressure ratio over
            the shock, p2/p1.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)
        theta = set_theta(theta)

        # *change* of variables
        betas = cls.beta(M1=M1, gamma=gamma, theta=theta)
        M1sinbeta = M1 * np.sin(betas)

        # Reuse normal shock relations
        ratio = NormalShock.p2_p1(M1=M1sinbeta, gamma=gamma)

        return tuple(ratio)

    @classmethod
    def rho2_rho1(cls, M1: Hint.nums, gamma: Hint.nums = None,
                  theta: Hint.nums = None) -> tuple:
        """
        Compute the downstream over upstream static density ratio.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.
            theta: The deflection angle of the downstream flow from upstream.
                Optional, defaults to 0 (recovering normal shock relation).

        Returns:
            weak shock solution, strong shock solution : Density ratio over the
            shock, rho2/rho1.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)
        theta = set_theta(theta)

        # *change* of variables
        betas = cls.beta(M1=M1, gamma=gamma, theta=theta)
        M1sinbeta = M1 * np.sin(betas)

        # Reuse normal shock relations
        ratio = NormalShock.rho2_rho1(M1=M1sinbeta, gamma=gamma)

        return tuple(ratio)

    @classmethod
    def T2_T1(cls, M1: Hint.nums, gamma: Hint.nums = None,
              theta: Hint.nums = None) -> tuple:
        """
        Compute the downstream over upstream static temperature ratio.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.
            theta: The deflection angle of the downstream flow from upstream.
                Optional, defaults to 0 (recovering normal shock relation).

        Returns:
            weak shock solution, strong shock solution : Temperature ratio over
            the shock, T2/T1.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)
        theta = set_theta(theta)

        # *change* of variables
        betas = cls.beta(M1=M1, gamma=gamma, theta=theta)
        M1sinbeta = M1 * np.sin(betas)

        # Reuse normal shock relations
        ratio = NormalShock.T2_T1(M1=M1sinbeta, gamma=gamma)

        return tuple(ratio)

    @classmethod
    def p02_p01(cls, M1: Hint.nums, gamma: Hint.nums = None,
                theta: Hint.nums = None) -> tuple:
        """
        Compute the downstream over upstream total pressure ratio.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.
            theta: The deflection angle of the downstream flow from upstream.
                Optional, defaults to 0 (recovering normal shock relation).

        Returns:
            weak shock solution, strong shock solution : Pressure ratio over
            the shock, pt2/pt1.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)
        theta = set_theta(theta)

        # *change* of variables
        betas = cls.beta(M1=M1, gamma=gamma, theta=theta)
        M1sinbeta = M1 * np.sin(betas)

        # Reuse normal shock relations
        ratio = NormalShock.p02_p01(M1=M1sinbeta, gamma=gamma)

        return tuple(ratio)

    @staticmethod
    def T02_T01(*args, **kwargs) -> tuple:
        """
        Compute the downstream over upstream total temperature ratio.

        Returns:
            weak shock solution, strong shock solution : Temperature ratio over
            the shock, Tt2/Tt1(=1).

        """
        NormalShock.T02_T01(*args, **kwargs)  # Warn if necessary
        return np.ones(1), np.ones(1)

    @classmethod
    def M2(cls, M1: Hint.nums, gamma: Hint.nums = None,
           theta: Hint.nums = None) -> tuple:
        """
        Compute the downstream Mach number.

        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.
            theta: The deflection angle of the downstream flow from upstream.
                Optional, defaults to 0 (recovering normal shock relation).

        Returns:
            weak shock solution, strong shock solution : Mach number after the
            shock, M2.

        """
        # Recast as necessary
        M1 = cast2numpy(M1)
        gamma = set_gamma(gamma)
        theta = set_theta(theta)

        # *change* of variables
        betas = cls.beta(M1=M1, gamma=gamma, theta=theta)
        M1sinbeta = M1 * np.sin(betas)

        # Reuse normal shock relations
        M2_with_fluff = NormalShock.M2(M1=M1sinbeta, gamma=gamma)
        M2 = M2_with_fluff / np.sin(np.array(betas) - theta)

        return tuple(M2)


class ExpansionFan(object):
    """
    Instantiable class providing access to properties of Prandtl-Meyer expansion
    fans.
    """

    def __init__(self, M1: Hint.nums, gamma: Hint.nums = None,
                 theta: Hint.nums = None):
        """
        Args:
            M1: Upstream Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.
            theta: The deflection angle of the downstream flow from upstream.
                Optional, defaults to 0 (no expansion).
        """
        # Recast as necessary
        self._M1 = cast2numpy(M1)
        self._gamma = set_gamma(gamma)
        self._theta = set_theta(theta)

        # Compute the Prandtl-Meyer angles before and after deflection
        nu1 = IsentropicFlow.nu(M=self._M1, gamma=self._gamma)
        nu2 = nu1 + self._theta

        # Cap the flow deflection angle as necessary
        nu_limit = nu_max(gamma=self._gamma)
        if any(nu2 > nu_limit):
            warnmsg = (
                "Maximum turn angle of supersonic flow has been reached, "
                "limiting turn angle of the flow as appropriate"
            )
            warnings.warn(message=warnmsg, category=RuntimeWarning)
            self._theta = np.where(nu2 > nu_limit, nu_limit - nu1, self._theta)

        # Create a solver that we can vectorize
        def solver(a, b):
            """
            Helper function to solve for Mach associated with nu.

            Args:
                a: Upstream gamma.
                b: Downstream Prandtl-Meyer angle.

            Returns:
                Mach number.

            """

            def f_opt(x):
                """Helper function for root-finding."""
                result = IsentropicFlow.nu(M=x, gamma=a) - b
                return result

            M = newton(f_opt, 2)
            return M

        # Map P-M angle after deflection to downstream Mach number
        self._M2 = np.vectorize(solver)(a=self._gamma, b=nu2)
        return

    def __repr__(self):
        reprstring = (
            f"{type(self).__name__}(M1={self._M1}, gamma={self._gamma}, "
            f"theta={self._theta})"
        )
        return reprstring

    @property
    def gamma(self) -> np.ndarray:
        """Flow adiabatic index."""
        return self._gamma

    @property
    def theta(self) -> np.ndarray:
        """Flow turning angle."""
        return self._theta

    @property
    def M1(self) -> np.ndarray:
        """Upstream Mach number."""
        return self._M1

    @property
    def M2(self) -> np.ndarray:
        """Downstream Mach number."""
        return self._M2

    @cached_property
    def T2_T1(self):
        """Static temperature ratio over the fan."""
        T2_Tt = IsentropicFlow.T_T0(M=self._M2, gamma=self._gamma)
        T1_Tt = IsentropicFlow.T_T0(M=self._M1, gamma=self._gamma)
        return T2_Tt / T1_Tt

    @property
    def p2_p1(self):
        """Static pressure ratio over the fan."""
        p2_p1 = self.T2_T1 ** (self._gamma / (self._gamma - 1))
        return p2_p1

    @property
    def rho2_rho1(self):
        """Static density ratio over the fan."""
        rho2_rho1 = self.T2_T1 ** (1 / (self._gamma - 1))
        return rho2_rho1


class RayleighFlow(object):
    """
    Classic Rayleigh flow methods for an ideal gas.
    """

    @staticmethod
    def p_pstar(M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of static pressure to static pressure at throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of static pressure to choked static pressure.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        ratio = (gamma + 1) / (1 + gamma * M ** 2)

        return ratio

    @classmethod
    def rho_rhostar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of static density to static density at throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of static density to choked static density.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        ratio = 1 / M ** 2 / cls.p_pstar(M=M, gamma=gamma)

        return ratio

    @classmethod
    def T_Tstar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of static temperature to static temperature at throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of static temperature to choked static temperature.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        ratio = (cls.p_pstar(M=M, gamma=gamma) * M) ** 2

        return ratio

    @classmethod
    def V_Vstar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of flow velocity to flow velocity at throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of flow velocity to choked flow velocity.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        ratio = 1 / cls.rho_rhostar(M=M, gamma=gamma)

        return ratio

    @classmethod
    def p0_p0star(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of stagnation pressure at a station to the throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of total pressures between a station and the throat.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)
        kwargs = {"M": M, "gamma": gamma}

        # Compute
        ratio = cls.p_pstar(**kwargs) / IsentropicFlow.p_pstar(**kwargs)

        return ratio

    @classmethod
    def T0_T0star(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of stagnation temperature at a station to the throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of given station and the throat total temperatures.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)
        kwargs = {"M": M, "gamma": gamma}

        # Compute
        ratio = cls.T_Tstar(**kwargs) / IsentropicFlow.T_Tstar(**kwargs)

        return ratio

    @classmethod
    def DeltaS(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the change in dimensionless entropy, deltaS.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The change in dimensionless entropy.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute the change in dimensionless entropy
        power = (gamma + 1) / gamma
        M = np.where(M <= 0, np.nan, M)
        DeltaS = np.log(M ** 2 * cls.p_pstar(M=M, gamma=gamma) ** power)

        return DeltaS

    @staticmethod
    def H(M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the dimensionless enthalpy, H.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The dimensionless enthalpy.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        sqrtH = (gamma + 1) * M / (1 + gamma * M ** 2)
        H = sqrtH ** 2

        return H


class FannoFlow(object):
    """
    Classic Fanno flow methods for an ideal gas.
    """

    @staticmethod
    def T_Tstar(M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of static temperature to static temperature at throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of static temperature to choked static temperature.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        ratio = (gamma + 1) / (2 + (gamma - 1) * M ** 2)

        return ratio

    @classmethod
    def p_pstar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of static pressure to static pressure at throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of static pressure to choked static pressure.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        ratio = cls.T_Tstar(M=M, gamma=gamma) ** 0.5 / M

        return ratio

    @classmethod
    def rho_rhostar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of static density to static density at throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of static density to choked static density.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        ratio = cls.T_Tstar(M=M, gamma=gamma) ** -0.5 / M

        return ratio

    @classmethod
    def V_Vstar(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of flow velocity to flow velocity at throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of flow velocity to choked flow velocity.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        ratio = cls.T_Tstar(M=M, gamma=gamma) ** 0.5 * M

        return ratio

    @staticmethod
    def T0_T0star(*args, **kwargs) -> np.ndarray:
        """
        Compute the ratio of stagnation temperature at a station to the throat.

        Returns:
            The ratio of given station and the throat total temperatures (= 1).

        """
        NormalShock.T02_T01(*args, **kwargs)  # Warn if necessary
        return np.ones(1)

    @classmethod
    def p0_p0star(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of stagnation temperature at a station to the throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of total temperatures between a station and the throat.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        power = -((gamma + 1) / (2 * (gamma - 1)))
        ratio = (cls.T_Tstar(M=M, gamma=gamma) ** power) / M

        return ratio

    @classmethod
    def rho0_rho0star(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the ratio of stagnation temperature at a station to the throat.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The ratio of total temperatures between a station and the throat.

        """
        return cls.p0_p0star(M=M, gamma=gamma)

    @staticmethod
    def H(M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the dimensionless enthalpy, H.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The dimensionless enthalpy.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute
        H = IsentropicFlow.T_T0(M=M, gamma=gamma)

        return H

    @classmethod
    def DeltaS(cls, M: Hint.nums, gamma: Hint.nums = None) -> np.ndarray:
        """
        Compute the change in dimensionless entropy, deltaS.

        Args:
            M: Mach number of the flow.
            gamma: Adiabatic index of the flow. Optional, defaults to 1.4.

        Returns:
            The change in dimensionless entropy.

        """
        # Recast as necessary
        M = cast2numpy(M)
        gamma = set_gamma(gamma)

        # Compute the change in dimensionless entropy
        power1 = (gamma - 1) / gamma
        power2 = (gamma + 1) / (2 * gamma)
        M = np.where(M <= 0, np.nan, M)
        DeltaS = np.log(M ** power1 * cls.T_Tstar(M=M, gamma=gamma) ** power2)

        return DeltaS


def visualise_flow():
    """Demonstrate relationships between various flow parameters."""
    fig, axs = plt.subplots(nrows=2, ncols=2, dpi=140, figsize=(9, 4.6),
                            sharex="all", sharey="all")
    fig.subplots_adjust(hspace=0.3, left=0.05, right=0.97, top=0.93, wspace=0.2)

    Machs = np.linspace(1e-6, 3, 100)

    # Isentropic flow
    axs.flat[0].set_title(IsentropicFlow.__name__)
    axs.flat[0].plot(Machs, IsentropicFlow.T_T0(Machs))
    axs.flat[0].plot(Machs, IsentropicFlow.p_p0(Machs))
    axs.flat[0].plot(Machs, IsentropicFlow.rho_rho0(Machs))
    axs.flat[0].plot(Machs, IsentropicFlow.A_Astar(Machs))

    # Normal shock jump
    axs.flat[1].set_title(NormalShock.__name__)
    axs.flat[1].plot(Machs, NormalShock.M2(Machs))
    axs.flat[1].plot(Machs, NormalShock.T2_T1(Machs))
    axs.flat[1].plot(Machs, NormalShock.p2_p1(Machs))
    axs.flat[1].plot(Machs, NormalShock.rho2_rho1(Machs))
    axs.flat[1].plot(Machs, NormalShock.p02_p01(Machs))

    # Rayleigh flow
    axs.flat[2].set_title(RayleighFlow.__name__)
    axs.flat[2].plot(Machs, RayleighFlow.T_Tstar(Machs))
    axs.flat[2].plot(Machs, RayleighFlow.p_pstar(Machs))
    axs.flat[2].plot(Machs, RayleighFlow.V_Vstar(Machs))
    axs.flat[2].plot(Machs, RayleighFlow.T0_T0star(Machs))
    axs.flat[2].plot(Machs, RayleighFlow.p0_p0star(Machs))

    # Fanno flow
    axs.flat[3].set_title(FannoFlow.__name__)
    axs.flat[3].plot(Machs, FannoFlow.T_Tstar(Machs))
    axs.flat[3].plot(Machs, FannoFlow.p_pstar(Machs))
    axs.flat[3].plot(Machs, FannoFlow.V_Vstar(Machs))
    axs.flat[3].plot(Machs, FannoFlow.p0_p0star(Machs))

    for i, ax in enumerate(axs.flat):
        if i > 1:
            ax.set_xlabel("Mach")
        ax.set_xlim(0, 3)
        ax.set_ylim(0, 3)

    plt.show()
    return None


if __name__ == "__main__":
    from matplotlib import pyplot as plt

    visualise_flow()



"""Test cases for aerodynamic methods relating to wing performance/geometry."""
import unittest

import numpy as np

from carpy.aerodynamics.aerofoil import NewAerofoil
from carpy.aerodynamics.wing import (
    WingSection, WingSections, PLLT, HorseshoeVortex, CDfGudmundsson
)  # , Cantilever1DStatic)


class Solvers(unittest.TestCase):
    """Check that methods relating to the generation of lift/drag results."""

    @staticmethod
    def wing_SuperLazarusMkII():
        fx76 = NewAerofoil.from_url(
            "https://m-selig.ae.illinois.edu/ads/coord/fx76mp140.dat")
        dae31 = NewAerofoil.from_url(
            "https://m-selig.ae.illinois.edu/ads/coord/dae31.dat")

        # Define buttock-line geometry
        mysections = WingSections()
        mysections[0] = WingSection(fx76)
        mysections[6] = WingSection(fx76)
        mysections[10] = WingSection(fx76)
        mysections[14] = WingSection(dae31)

        # Add sweep, dihedral, and twist
        mysections[:6].sweep = np.radians(0)
        # mysections[6:].sweep = np.radians(2)
        mysections[6:].sweep = np.radians(0)
        mysections[0:].dihedral = np.radians(0)
        mysections[0:].twist = np.radians(0)
        mysections[14].twist = np.radians(-3)

        # Introduce wing taper
        mysections[0:].chord = 1.0
        mysections[14].chord = 0.3
        return mysections

    def test_running(self):
        """Check that the methods are even capable of running."""
        mysections = self.wing_SuperLazarusMkII()

        aoa = np.radians(3)
        soln0 = PLLT(sections=mysections, span=24, alpha=aoa)
        soln1 = HorseshoeVortex(sections=mysections, span=24, alpha=aoa)
        # soln2 = Cantilever1DStatic(
        #     sections=mysections, spar=None, span=24, alpha=aoa,
        #     lift=115 * 9.81, N=60
        # )
        return

    def test_naca0012elevator(self):
        """Compare results against XFLR5 test case."""

        n0012 = NewAerofoil.from_method.NACA("0012")

        mysections = WingSections()
        mysections[0] = WingSection(n0012)
        mysections[100] = WingSection(n0012)

        mysections[:].sweep = 0
        mysections[:].dihedral = 0
        mysections[:].twist = 0

        mysections[0].chord = 0.65
        mysections[100].chord = 0.24

        aoa = np.radians(10)
        soln0 = PLLT(sections=mysections, span=4, alpha=aoa)
        soln1 = HorseshoeVortex(sections=mysections, span=4, alpha=aoa)

        return


class GudmundssonSkinFriction(unittest.TestCase):

    def test_method(self):
        n2412 = NewAerofoil.from_method.NACA("2412")

        mysections = WingSections()
        mysections[0] = WingSection(n2412)
        mysections[100] = WingSection(n2412)

        from carpy.utility import Quantity
        mysections[0].chord = Quantity(4.88, "ft")
        mysections[100].chord = Quantity(2.59, "ft")

        mysections[:].sweep = 0
        mysections[:].dihedral = 0
        mysections[:].twist = 0

        soln = CDfGudmundsson(
            sections=mysections, span=Quantity(38.3, "ft"),
            altitude=0, TAS=Quantity(185, "kt"))

        self.assertTrue(np.isclose(soln._Cf, 0.001999, atol=1e-4))

        return
